#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Context for a data analysis environment.

Provides a context class, its setup and an attachment function
"""

DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(asctime)s %(levelname)s %(filename)s:%(lineno)d (%(funcName)s): %(message)s"
        }
    },
    "handlers": {
        "screen": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        },
        "file": {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "default",
            "filename": "debug.log",
            "mode": "a",
        },
    },
    "loggers": {"": {"level": "DEBUG", "handlers": ["screen", "file"]}},
}

import ast
import configparser
import logging
import logging.config
import os

from rich import traceback
from typeguard import typechecked

traceback.install()
log = logging.getLogger(__name__)


class Context:
    """
    A class used to store various paths and properties.

    This may be reduced to being a dictionary in the future.
    """

    def __init__(self) -> None:
        self.paths = {}
        self.prep = {}
        self.cores = 1
        self.max_kvsize = 1099511627776  # 1 TB

    @typechecked
    def gather_paths(self, config: configparser.ConfigParser, defaults: dict) -> None:
        """
        Gathers working paths from config file and stores them in self.

        Args:
            config (ConfigParser): The configuration object.
        """
        self.paths["work"] = os.path.join(
            config.get(
                "storagetools.locations",
                "workdir_root",
                fallback=defaults["storagetools.locations"]["workdir_root"],
            ),
            config.get(
                "environment",
                "analysis_name",
                fallback=defaults["environment"]["analysis_name"],
            ),
        )
        self.paths["storage"] = os.path.join(
            self.paths["work"],
            config.get(
                "storagetools.locations",
                "storage_dir",
                fallback=defaults["storagetools.locations"]["storage_dir"],
            ),
        )
        self.paths["database"] = os.path.join(
            self.paths["work"],
            config.get(
                "storagetools.locations",
                "database_filename",
                fallback=defaults["storagetools.locations"]["database_filename"],
            ),
        )
        log.debug("paths gathered from config")

    @typechecked
    def paths_prepend(self, path_to_main: str) -> None:
        """
        Prepends the main directory path to the existing paths.

        Args:
            path_to_main (str): The path to the main directory.
        """
        for k, v in self.paths.items():
            self.paths[k] = os.path.join(path_to_main, v)
        log.info(f"prepended '{path_to_main}' to ctx.paths")

    @typechecked
    def gather_property_columns(
        self, config: configparser.ConfigParser, defaults: dict
    ) -> None:
        """
        Gathers the property columns from the config file.

        This affects experiments, sensors and channels.

        Args:
            config (ConfigParser): The configuration object.
        """
        self.experiment_properties = ast.literal_eval(
            config.get(
                "property_columns",
                "experiment",
                fallback=str(defaults["property_columns"]["experiment"]),
            )
        )
        self.sensor_properties = ast.literal_eval(
            config.get(
                "property_columns",
                "sensor",
                fallback=str(defaults["property_columns"]["sensor"]),
            )
        )
        self.channel_properties = ast.literal_eval(
            config.get(
                "property_columns",
                "channel",
                fallback=str(defaults["property_columns"]["channel"]),
            )
        )
        log.debug("properties gathered from config")

    def create_folders(self):
        """
        Creates necessary folders (workdir).
        """
        os.makedirs(self.paths["work"], exist_ok=True)
        os.makedirs(self.paths["storage"], exist_ok=True)
        log.debug("folders created")


@typechecked
def attach(conf="config.ini", path_to_main: str = ".") -> Context:
    """
    Attaches to a context as described in the configuration.
    Set up logging and working paths.

    Args:
        conf (str): The path to the configuration file. Defaults to "config.ini".
        path_to_main (str, optional): The path to the main directory. Defaults
            to '.', meaning the config is stored in same directory. This is
            relevant, if a data analysis project contains many scripts in
            different subdirectories and the config file is stored in the top
            level.

    Returns:
        Context: The context object with the gathered paths and properties.
    """
    # Define default values
    defaults = {
        "environment": {
            "analysis_name": "default_analysis",
            "available_cores": 1,
            "max_kvsize": 1099511627776,
        },
        "storagetools.locations": {
            "workdir_root": "work",
            "storage_dir": "data",
            "plot_dir": "plots",
            "database_filename": "database.db",
        },
        "property_columns": {"experiment": {}, "sensor": {}, "channel": {}},
    }

    if path_to_main != ".":
        conf = os.path.join(path_to_main, conf)
    if not os.path.exists(conf):
        with open(conf, "w") as f:
            f.write("[environment]\nanalysis_name = default_analysis")
    try:
        logging.config.fileConfig(conf, disable_existing_loggers=False)
    except KeyError:
        logging.config.dictConfig(DEFAULT_LOGGING_CONFIG)

    config = configparser.ConfigParser()
    config.read(conf)
    log.info(
        f"{config.get('environment', 'analysis_name', fallback=defaults['environment']['analysis_name'])}"
    )

    ctx = Context()
    ctx.gather_paths(config, defaults)
    # Prepend main path if necessary
    if path_to_main != ".":
        ctx.paths_prepend(path_to_main)
    ctx.gather_property_columns(config, defaults)
    ctx.create_folders()

    # Writes maximum size for LMDB key value storage from config file to ctx.
    ctx.max_kvsize = int(
        config.get(
            "environment", "max_kvsize", fallback=defaults["environment"]["max_kvsize"]
        )
    )
    log.debug(f"set max_kvsize to {ctx.max_kvsize}")

    # Get available cores with default
    ctx.cores = int(
        config.get(
            "environment",
            "available_cores",
            fallback=defaults["environment"]["available_cores"],
        )
    )
    log.debug(f"set available_cores to {ctx.cores}")

    return ctx



#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Machine learning related classes for data analysis."""

import logging
from typing import Any

from storagetools import helpers
from storagetools.context import Context
from storagetools import model
from storagetools.model import Identifiable, Linkable
from typeguard import typechecked

log = logging.getLogger(__name__)


class Datapoint(Linkable):
    """
    A class for an Input-Output mapping, intended for complex Inputs.

    has no serialization possibility at the moment
    """

    @typechecked
    def __init__(self, X: Identifiable, y: Any) -> None:
        """
        Args:
            X: an identifiable (TSRepresentation or Container or ...)
            y: label information (str or int or ...)
        """

        Linkable.__init__(self)
        self.X = X
        self.y = y


class Dataset(Linkable, dict):
    """
    A convenience class for organizations of multiple complex datapoint.

    Is supposed to store X and y, as well as others, for example:
        X_featurized
        X_transformed
        y_pred
        y_encoded

    Has no serialization possibility at the moment.
    """

    @classmethod
    @typechecked
    def from_datapoints_container(
        cls, ctx: Context, container: model.Container
    ) -> "Dataset":
        """
        creates Dataset from container (container may contain ONLY datapoints)

        so far no check function implemented
        """
        if container.lading == []:
            helpers.deserialize_lading(ctx, container)

        X = [datapoint.X for datapoint in container.lading]
        y = [datapoint.y for datapoint in container.lading]
        return cls(X, y)

    @typechecked
    def __init__(self, X: list, y: list = [], *args, **kwargs) -> None:
        """
        Args:
            X: a list of datapoints
            y: a list of label information (same order as X)
            args, kwargs: allows for enrichment with other information
        """

        Linkable.__init__(self)
        dict.__init__(self, *args, **kwargs)
        self["X"] = X
        self["y"] = y

    def as_dict(self) -> dict:
        """
        An export function, to share for use without storagetools.

        Returns:
            dict representation of the dataset
        """
        result = {}
        for k, v in self.items():
            result[k] = v
        return result



#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Manages (de-) serialization

This module provides functions for serializing and deserializing objects to and
from a SQLite database and an LMDB key-value store.

It includes classes for serializing and deserializing objects, as well as
functions for removing objects from the database.
"""

from typing import Any
import lmdb
import pickle
import logging
import sys
import warnings
import uuid

from tqdm import tqdm
from typeguard import typechecked

from storagetools.context import Context
from storagetools.database import Database
from storagetools.model import (
    Channel,
    ChannelCapture,
    Container,
    Experiment,
    HasPropertiesFromCase,
    Identifiable,
    IsCapture,
    Linkable,
    Sensor,
    SensorCapture,
    TSRepresentation,
)

warnings.filterwarnings("ignore", message="metadata on a dtype")

log = logging.getLogger(__name__)

CLASSNAME_TO_TABLE_MAPPING = {
    "Experiment": "experiment",
    "Sensor": "sensor",
    "Channel": "channel",
    "SensorCapture": "sensor_capture",
    "ChannelCapture": "channel_capture",
    "TSRepresentation": "ts_representation",
    "Container": "container",
}

CLASS_TO_TABLE_MAPPING = {
    Experiment: "experiment",
    Sensor: "sensor",
    Channel: "channel",
    SensorCapture: "sensor_capture",
    ChannelCapture: "channel_capture",
    TSRepresentation: "ts_representation",
    Container: "container",
}


class Serializer:
    """
    A class for serializing objects to the metadata database and LMDB key-value store.
    """

    class_to_table_mapping = CLASS_TO_TABLE_MAPPING

    def __init__(self, target: Any) -> None:
        self.target = target
        self.references = []

    def generate_record_and_sql(self) -> tuple[str, dict]:
        """
        Generates a record and SQL statement for inserting the object into the database.

        Returns:
            A tuple containing the SQL statement and the record.
        """
        record = {}
        from_supers = self.gather_record_entries_from_supers()
        from_instance = self.gather_record_entries_from_instance()
        record.update(from_supers)
        record.update(from_instance)
        sql = self.assemble_sql_string(record)
        return sql, record

    def gather_record_entries_from_supers(self) -> dict:
        """
        collect entries from inherited fields (Identifiable, Linkable, ...)

        Returns:
            A dictionary with entries
        """
        record = {}
        if issubclass(self.target.__class__, Identifiable):
            record["guid"] = self.target.guid

        if issubclass(self.target.__class__, Linkable):
            for association in self.target.associations:
                self.references.append((self.target.guid, association))
            for attachment in self.target.attachments:
                self.references.append((attachment, self.target.guid))

        if issubclass(self.target.__class__, HasPropertiesFromCase):
            record.update(self.target.properties)

        if issubclass(self.target.__class__, IsCapture):
            record["capture_of"] = self.target.capture_of
        return record

    def gather_record_entries_from_instance(self) -> dict:
        """
        Collects entries from instance fields (TSRepresentation, Container, ...)

        Returns:
            A dictionary with entries
        """
        record = {}
        if issubclass(self.target.__class__, TSRepresentation):
            record["name"] = self.target.name
            record["type"] = self.target.type
        if issubclass(self.target.__class__, Container):
            record["name"] = self.target.name
            record["info"] = self.target.info
        return record

    def which_table(self) -> str:
        """
        Returns the name of the table in the database that corresponds to the object's class.

        Returns:
            str: name of table
        """
        return self.class_to_table_mapping[self.target.__class__]

    @typechecked
    def assemble_sql_string(self, record: dict) -> str:
        """
        Assembles an SQL statement for inserting the object into the database.

        Args:
            record: The record containing the object's metadata.

        Returns:
            The SQL statement.
        """
        sql = f"""
        INSERT OR REPLACE INTO {self.which_table()}
        {tuple(record.keys())}
        VALUES (?{(len(record.keys()) - 1) * ", ?"})
        """
        return sql

    @typechecked
    def store_blob(self, ctx: Context) -> None:
        """
        Stores the object's payload in the LMDB key-value store.

        Args:
            ctx: The context object containing the paths to the LMDB key-value store.
        """
        kvs = lmdb.open(ctx.paths["storage"], map_size=ctx.max_kvsize)
        payload = pickle.dumps(self.target.payload)
        with kvs.begin(write=True) as txn:
            txn.put(self.target.guid.bytes, payload)
        kvs.close()

    @typechecked
    def serialize_core(self, ctx: Context) -> None:
        """
        Serializes the object's core (metadata) data to the database.

        Should be renamed to serialize_metadata or sth in the future.

        Args:
            ctx: The context object containing the paths to the database.
        """
        sql, record = self.generate_record_and_sql()
        with Database(ctx) as db:
            db.cur.execute(sql, list(record.values()))

    @typechecked
    def serialize_references(self, ctx: Context) -> None:
        """
        Serializes the object's references to other objects to the database.

        References means every association (reference 'up') or attachment (reference 'down').

        Args:
            ctx: The context object containing the paths to the database.
        """
        sql = """
        INSERT OR REPLACE INTO association (child, parent) VALUES (?, ?)
        """
        with Database(ctx) as db:
            db.cur.executemany(sql, self.references)


@typechecked
def serialize(ctx: Context, target: Any, blob_also: bool = True) -> None:
    """
    Serializes an object.

    Serializes an object to the metadata database and optionally its payload to the LMDB key-value store.

    Args:
        ctx: The context object containing the paths to the database and LMDB key-value store.
        target: The object to serialize.
        blob_also: Whether to serialize the object payload to the LMDB key-value store. Defaults to True.
    """
    serializer = Serializer(target)
    serializer.serialize_core(ctx)
    if issubclass(target.__class__, Linkable):
        serializer.serialize_references(ctx)
    if blob_also:  # can be skipped if for example only association is updated
        if issubclass(target.__class__, TSRepresentation):
            serializer.store_blob(ctx)


@typechecked
def serialize_list(ctx: Context, target_list: list[Any]) -> None:
    """
    Convenience function that serializes a list of objects to database key-value store.

    Args:
        ctx: The context object containing the paths to the database and LMDB key-value store.
        target_list: The list of objects to serialize.
    """
    kvs = lmdb.open(ctx.paths["storage"], map_size=ctx.max_kvsize)
    with kvs.begin(write=True) as txn:
        for target in tqdm(target_list):
            serialize(ctx, target, blob_also=False)
            if issubclass(target.__class__, TSRepresentation):
                payload = pickle.dumps(target.payload)
                txn.put(target.guid.bytes, payload)
    kvs.close()


class Deserializer:
    """
    A class for deserializing objects from the metadata database and LMDB key-value store.
    """

    class_to_table_mapping = CLASS_TO_TABLE_MAPPING

    def __init__(self, target_guid: uuid.UUID) -> None:
        self.target_guid = target_guid
        self.record = {}

    @typechecked
    def get_class(self, ctx: Context) -> None:
        """
        Gets the class of the object the deserializer was initialized with.

        The GUID could resolve to anything, thus its class has to be determined
        in order to find the table in the database from which to deserialize
        metadata from.

        Args:
            ctx: The context object containing the paths to the database and
                LMDB key-value store.
        """
        sql = "SELECT tablename FROM guid_lookup WHERE guid = ?"
        with Database(ctx) as db:
            tablename = db.cur.execute(sql, (self.target_guid,)).fetchone()[0]
        self.class_ = [
            key
            for key, value in self.class_to_table_mapping.items()
            if value == tablename
        ][0]

    @typechecked
    def get_record_entries_from_supers(self, ctx: Context) -> dict:
        """
        Collects entries from inherited fields (Linkable, HasPropertiesFromCase, IsCapture)

        Args:
            ctx: The context object containing the paths to the database and
                LMDB key-value store.

        Returns:
            A dict with metadata
        """
        record = {}
        table = self.class_to_table_mapping[self.class_]

        if issubclass(self.class_, Linkable):
            record["associations"] = []
            record["attachments"] = []
            # fetch guid from associations where guid = parent or = child
            sql = "SELECT parent FROM association WHERE child = ?"
            with Database(ctx) as db:
                results = db.cur.execute(sql, (self.target_guid,)).fetchall()
            for result in results:
                record["associations"].append(result[0])

            sql = "SELECT child FROM association WHERE parent = ?"
            with Database(ctx) as db:
                results = db.cur.execute(sql, (self.target_guid,)).fetchall()
            for result in results:
                record["attachments"].append(result[0])

        if issubclass(self.class_, HasPropertiesFromCase):
            record["properties"] = {}
            if self.class_ == Experiment:
                property_set = ctx.experiment_properties
            elif self.class_ == Sensor:
                property_set = ctx.sensor_properties
            elif self.class_ == Channel:
                property_set = ctx.channel_properties
            else:
                log.warning("property set not defined for deserialization!")
                property_set = {}

            # iterate over types (_) in properties (integer, text, etc)
            for _, cols in property_set.items():
                sql = f"""
                SELECT {", ".join(cols)} FROM {table} WHERE guid = ?
                """
                with Database(ctx) as db:
                    vals = db.cur.execute(sql, (self.target_guid,)).fetchone()
                result = {}
                for i, col in enumerate(cols):
                    result[col] = vals[i]
                record["properties"].update(result)

        if issubclass(self.class_, IsCapture):
            sql = f"SELECT capture_of FROM {table} WHERE guid = ?"
            with Database(ctx) as db:
                record["capture_of"] = db.cur.execute(
                    sql, (self.target_guid,)
                ).fetchone()[0]
        return record

    @typechecked
    def get_record_entries_from_instance(self, ctx: Context) -> dict:
        """
        Collects entries specific to distinct classes (TSRepresentation, Container)

        May be merged with get_record_entries_from_supers in the future, works similarly.

        Args:
            ctx: The context object containing the paths to the database and
                LMDB key-value store.

        Returns:
            A dict with metadata
        """

        record = {}
        table = self.class_to_table_mapping[self.class_]

        if issubclass(self.class_, TSRepresentation):
            sql = f"SELECT name, type FROM {table} WHERE guid = ? "
            with Database(ctx) as db:
                result = db.cur.execute(sql, (self.target_guid,)).fetchone()
            record["name"] = result[0]
            record["type"] = result[1]

        if issubclass(self.class_, Container):
            sql = f"SELECT name, info FROM {table} WHERE guid = ? "
            with Database(ctx) as db:
                result = db.cur.execute(sql, (self.target_guid,)).fetchone()
            record["name"] = result[0]
            record["info"] = result[1]
        return record

    @typechecked
    def restore_blob(self, ctx: Context) -> Any:
        """
        Retrieves payload data from key value store.

        Args:
            ctx: The context object containing the paths to the database and
                LMDB key-value store.

        Returns:
            Whatever payload is related to GUID
        """
        kvs = lmdb.open(ctx.paths["storage"], map_size=ctx.max_kvsize)
        with kvs.begin() as txn:
            payload = pickle.loads(txn.get(self.target_guid.bytes))
        kvs.close()
        return payload

    @typechecked
    def delete_blob(self, ctx: Context) -> None:
        """
        deletes payload data related to deserializer guid from key value store.

        args:
            ctx: the context object containing the paths to the database and
                lmdb key-value store.
        """
        kvs = lmdb.open(ctx.paths["storage"], map_size=ctx.max_kvsize)
        with kvs.begin(write=True) as txn:
            txn.delete(self.target_guid.bytes)
        kvs.close()

    @typechecked
    def create_from_record(self, ctx: Context, record: dict) -> Any:
        """
        Recreates an object from the collected metadata.

        Requires class of the target object to be determined and metadata collected.

        args:
            ctx: the context object containing the paths to the database and
                lmdb key-value store.
            record: dict with metadata

        Returns:
            desieralized object
        """

        if self.class_ == Experiment:
            x = Experiment(
                properties=record["properties"],
            )
            x.guid = self.target_guid
        elif self.class_ == Sensor:
            x = Sensor(
                properties=record["properties"],
            )
            x.guid = self.target_guid
        elif self.class_ == Channel:
            x = Channel(
                properties=record["properties"],
            )
            x.guid = self.target_guid
            x.properties = record["properties"]
        elif self.class_ == SensorCapture:
            x = SensorCapture(capture_of=record["capture_of"])
            x.guid = self.target_guid
        elif self.class_ == ChannelCapture:
            x = ChannelCapture(capture_of=record["capture_of"])
            x.guid = self.target_guid
        elif self.class_ == Container:
            x = Container()
            x.guid = self.target_guid
            x.name = record["name"]
            x.info = record["info"]
        elif self.class_ == TSRepresentation:
            payload = self.restore_blob(ctx)
            x = TSRepresentation(
                payload=payload,
                name=record["name"],
                type_=record["type"],
            )
            x.guid = self.target_guid
        else:
            log.error(f"no matching class for {self.class_}")
            sys.exit()

        if issubclass(self.class_, Linkable):
            x.attachments = record["attachments"]  # type: ignore
            x.associations = record["associations"]  # type: ignore
        return x

    @typechecked
    def core_remove_from_db(self, ctx: Context) -> None:
        """
        Removes an objects metadata.

        Think twice, maybe use db_interface.remove().
        Does not touch the key value store.
        May be renamed in the future and merged with delete_blob.

        Args:
            ctx: the context object containing the paths to the database and
                lmdb key-value store.
        """

        self.get_class(ctx)
        table = self.class_to_table_mapping[self.class_]

        # remove row from table
        sql = f"DELETE FROM {table} WHERE guid = ?"
        with Database(ctx) as db:
            db.cur.execute(sql, (self.target_guid,))

        # remove links
        sql = """
        DELETE FROM association 
        WHERE parent = ? 
        OR child = ? 
        """
        with Database(ctx) as db:
            db.cur.execute(sql, (self.target_guid, self.target_guid))


@typechecked
def deserialize(ctx: Context, target_guid: uuid.UUID) -> Any:
    """
    Deserializes an object from the database and LMDB key-value store.

    Args:
        ctx: The context object containing the paths to the database and LMDB
            key-value store.
        target_guid: The GUID of the object to deserialize
    Returns:
        The deserialized object.
    """
    deserializer = Deserializer(target_guid)
    deserializer.get_class(ctx)
    record = {}
    record.update(deserializer.get_record_entries_from_supers(ctx))  # type: ignore
    record.update(deserializer.get_record_entries_from_instance(ctx))
    x = deserializer.create_from_record(ctx, record)
    return x


@typechecked
def deserialize_guid_list(
    ctx: Context,
    guid_list: list[uuid.UUID],
    disable_progressbar: bool = True,
) -> list[Any]:
    """
    convenience function, does what it says

    Args:
        ctx: The context object containing the paths to the database and LMDB
            key-value store.
        guid_list: list of guids to deserialize
        progressbar: tqdm progressbar
    Returns:
        list of deserialized objects
    """
    x = []
    for guid in tqdm(guid_list, disable=disable_progressbar):
        x.append(deserialize(ctx, guid))
    return x


@typechecked
def remove(
    ctx: Context,
    target_guid: uuid.UUID,
    blob_also: bool = True,
) -> None:
    """
    Removes object from metadata database, also from key-value store.

    Args:
        ctx: The context object containing the paths to the database and LMDB
            key-value store.
        target_guid: GUID of object to be removed
        blob_also (optional): Whether to remove the blob also (otherwise metadata only)
    """
    deserializer = Deserializer(target_guid)
    deserializer.core_remove_from_db(ctx)
    if blob_also:
        if issubclass(deserializer.class_, TSRepresentation):
            deserializer.delete_blob(ctx)


@typechecked
def remove_blob_only(ctx: Context, target_guid: uuid.UUID) -> None:
    """
    Removes object from key-value store only.

    Args:
        ctx: The context object containing the paths to the database and LMDB
            key-value store.
        target_guid: GUID of object to be removed
    """
    deserializer = Deserializer(target_guid)
    deserializer.delete_blob(ctx)



#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Metadata database setup.

This module provides classes and functions for working with a SQLite metadata database.

It includes classes for working an in-memory or file-based representation of
the SQLite database, as well as functions for initializing the database,
creating tables and views, and adding property columns to the database.
"""

import logging
import os
import sqlite3
import uuid

from storagetools.context import Context
from typeguard import typechecked

DB_SCHEMA_FILE = "database_schema.sql"
DB_LOOKUP_VIEW_FILE = "database_lookup_view.sql"

log = logging.getLogger(__name__)


class Database:
    """
    A class for working with a SQLite (sqlite3) database.

    The class provides a context manager for working with the database,
    and provides methods for executing SQL queries.

    Note on UUIDs:
    UUIDs are not natively supported by SQLite. String representation requires
    32-36 bytes storage depending on notation, whereas a binary representation
    only 16 bytes. GUID columns are thus registered as 'memoryview' to keep
    metadata databases smaller.
    """

    @typechecked
    def __init__(self, ctx: Context) -> None:
        sqlite3.register_converter("GUID", lambda b: uuid.UUID(bytes=b))
        sqlite3.register_adapter(uuid.UUID, lambda u: memoryview(u.bytes))
        self.con = sqlite3.connect(
            f"file:{ctx.paths['database']}",
            detect_types=sqlite3.PARSE_DECLTYPES,
            uri=True,
        )
        self.cur = self.con.cursor()

    def __enter__(self) -> "Database":
        return self

    def __exit__(self, ext_type, exc_value, traceback) -> None:
        """
        Type hints missing here, also unused parameters.
        To be checked how this is implemented the deeper backend for proper resolution.
        """
        self.cur.close()
        if isinstance(exc_value, Exception):
            self.con.rollback()
        else:
            self.con.commit()
        self.con.close()


@typechecked
def initialize(ctx: Context) -> None:
    """
    Initializes the database by creating tables, adding property columns,
    and creating views.

    Args:
        ctx: The context object containing the paths to the database.
    """
    create_tables(ctx)
    add_property_columns(ctx)
    create_views_from_file(ctx)
    create_cc_repr_view(ctx)
    create_cc_view(ctx)
    create_sc_repr_view(ctx)


@typechecked
def create_tables(ctx: Context) -> None:
    """
    Creates tables in the database by executing SQL commands from a file.

    Args:
        ctx: The context object containing the paths to the database.
    """
    module_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.join(module_dir, "sql_scripts", DB_SCHEMA_FILE)
    with open(file_path, "r") as f:
        sql_script = f.read()
    with Database(ctx) as db:
        db.cur.executescript(sql_script)
    log.info("Empty tables created")


@typechecked
def create_views_from_file(ctx: Context) -> None:
    """
    Creates views in the database by executing SQL commands from a file.

    Args:
        ctx: The context object containing the paths to the database.
    """
    module_dir = os.path.dirname(os.path.abspath(__file__))
    file_paths = []
    file_paths.append(os.path.join(module_dir, "sql_scripts", DB_LOOKUP_VIEW_FILE))
    for file_path in file_paths:
        with open(file_path, "r") as f:
            sql_script = f.read()
        with Database(ctx) as db:
            db.cur.executescript(sql_script)
        log.info(f"created view from {file_path}")


@typechecked
def create_cc_repr_view(ctx: Context) -> None:
    """
    Creates a view for all representations attached to a channel capture.

    Some scenarios make use of reprs that are attached to sensorcaptures
    directly (e.g. channel averages) or not to anything at all. Those reprs
    will not be viewable in the cc_repr_view. Dealing with partly overlapping
    columns and merging them into one view got so complicated that it was
    overthrown as idea.

    See also: create_sc_repr_view(ctx)

    Args:
        ctx: The context object containing the property columns for the experiment, sensor, and channel.
    """
    tmp = []
    for _, items in ctx.channel_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    channel_properties = ", ".join(
        f'channel."{item}" as channel_{item}' for item in tmp
    )
    log.info(channel_properties)

    tmp = []
    for _, items in ctx.sensor_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    sensor_properties = ", ".join(f'sensor."{item}" as sensor_{item}' for item in tmp)
    log.info(sensor_properties)

    tmp = []
    for _, items in ctx.experiment_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    experiment_properties = ", ".join(
        f'experiment."{item}" as experiment_{item}' for item in tmp
    )
    log.info(experiment_properties)

    sql = f"""
    CREATE VIEW cc_repr_view AS
    SELECT
    ts_representation."guid" as ts_representation_guid,
    ts_representation."name" as ts_representation_name,
    ts_representation."type" as ts_representation_type,
    channel."guid" as channel_guid,
    {channel_properties}{"," if channel_properties else ""}
    cc."guid" as channelcapture_guid,
    sensor."guid" as sensor_guid,
    {sensor_properties}{"," if sensor_properties else ""}
    sc."guid" as sensorcapture_guid,
    experiment."guid" as experiment_guid{"," if experiment_properties else ""}
    {experiment_properties}
    FROM ts_representation
    JOIN association a1 ON ts_representation.guid = a1.child
    JOIN channel_capture cc ON a1.parent = cc.guid
    JOIN association a11 ON cc.guid = a11.child
    JOIN sensor_capture sc ON a11.parent = sc.guid
    JOIN association a12 ON sc.guid = a12.child
    JOIN experiment ON a12.parent = experiment.guid
    JOIN channel ON cc.capture_of = channel.guid
    JOIN association a2 ON channel.guid = a2.child
    JOIN sensor ON a2.parent = sensor.guid
    """

    log.debug(sql)
    with Database(ctx) as db:
        db.cur.execute(sql)


@typechecked
def create_cc_view(ctx: Context) -> None:
    """
    Creates a view for all channel captures.

    almost clone of create_cc_repr_view
    See also: create_cc_repr_view(ctx)

    Args:
        ctx: The context object containing the property columns for the experiment, sensor, and channel.
    """
    tmp = []
    for _, items in ctx.channel_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    channel_properties = ", ".join(
        f'channel."{item}" as channel_{item}' for item in tmp
    )
    log.info(channel_properties)

    tmp = []
    for _, items in ctx.sensor_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    sensor_properties = ", ".join(f'sensor."{item}" as sensor_{item}' for item in tmp)
    log.info(sensor_properties)

    tmp = []
    for _, items in ctx.experiment_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    experiment_properties = ", ".join(
        f'experiment."{item}" as experiment_{item}' for item in tmp
    )
    log.info(experiment_properties)

    sql = f"""
    CREATE VIEW cc_view AS
    SELECT
    channel."guid" as channel_guid,
    {channel_properties}{"," if channel_properties else ""}
    cc."guid" as channelcapture_guid,
    sensor."guid" as sensor_guid,
    {sensor_properties}{"," if sensor_properties else ""}
    sc."guid" as sensorcapture_guid,
    experiment."guid" as experiment_guid{"," if experiment_properties else ""} 
    {experiment_properties}
    FROM channel_capture cc
    JOIN association a11 ON cc.guid = a11.child
    JOIN sensor_capture sc ON a11.parent = sc.guid
    JOIN association a12 ON sc.guid = a12.child
    JOIN experiment ON a12.parent = experiment.guid
    JOIN channel ON cc.capture_of = channel.guid
    JOIN association a2 ON channel.guid = a2.child
    JOIN sensor ON a2.parent = sensor.guid
    """

    log.debug(sql)
    with Database(ctx) as db:
        db.cur.execute(sql)


@typechecked
def create_sc_repr_view(ctx: Context) -> None:
    """
    Creates a view for all representations attached to a sensor capture.

    code very similar to create_cc_repr_view
    See also: create_cc_repr_view(ctx)

    Args:
        ctx: The context object containing the property columns for the experiment and sensor.
    """
    tmp = []
    for _, items in ctx.sensor_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    sensor_properties = ", ".join(f'sensor."{item}" as sensor_{item}' for item in tmp)

    tmp = []
    for _, items in ctx.experiment_properties.items():
        for item in items:  # type: ignore
            tmp.append(item)
    experiment_properties = ", ".join(
        f'experiment."{item}" as experiment_{item}' for item in tmp
    )

    sql = f"""
    CREATE VIEW sc_repr_view AS
    SELECT
    ts_representation."guid" as ts_representation_guid,
    ts_representation."name" as ts_representation_name,
    ts_representation."type" as ts_representation_type,
    sensor."guid" as sensor_guid, 
    {sensor_properties}{"," if sensor_properties else ""}
    sc."guid" as sensorcapture_guid,
    experiment."guid" as experiment_guid{"," if experiment_properties else ""}
    {experiment_properties}
    FROM ts_representation
    JOIN association a1 ON ts_representation.guid = a1.child
    JOIN sensor_capture sc ON a1.parent = sc.guid
    JOIN association a12 ON sc.guid = a12.child
    JOIN experiment ON a12.parent = experiment.guid
    JOIN sensor ON sc.capture_of = sensor.guid
    """

    log.debug(sql)
    with Database(ctx) as db:
        db.cur.execute(sql)


@typechecked
def add_cols_to_table(
    ctx: Context,
    table: str,
    cols: list[str],
    coltype: str = "TEXT",
) -> None:
    """
    Adds columns to a table in the database.

    Args:
        ctx: The context object containing the paths to the database.
        table: The name of the table to add columns to.
        cols: A list of column names to add to the table.
        coltype: The data type of the columns to add. Defaults to "TEXT". See
            SQLite types for other options.
    """
    with Database(ctx) as db:
        for col in cols:
            sql = f"ALTER TABLE {table} ADD COLUMN {col} {coltype}"
            db.cur.execute(sql)
        db.con.commit()
    log.info(f"Altered table {table} by appending cols {cols}")


@typechecked
def add_property_columns(ctx: Context) -> None:
    """
    Adds property columns to experiment, sensor, and channel tables in DB.

    Args:
        ctx: The context object containing the property columns for the
            experiment, sensor, and channel.
    """
    for coltype, columns in ctx.experiment_properties.items():
        add_cols_to_table(ctx, "experiment", columns, coltype)
    for coltype, columns in ctx.sensor_properties.items():
        add_cols_to_table(ctx, "sensor", columns, coltype)
    for coltype, columns in ctx.channel_properties.items():
        add_cols_to_table(ctx, "channel", columns, coltype)


class MemoryDB:
    """
    A class for working with an in-memory SQLite database.

    The class creates an in-memory database and copies the contents of a
    file-based database into it. The file-based database is then closed, and
    the in-memory database is kept open. The class also provides a method for
    updating the context object with the location of the in-memory database.
    When the in-memory DB is closed, changes are written to the file-based
    origin.


    # Usage example

    ctx = context.attach()
    memory_db = database.use_inmemory(ctx)
    ctx = memory_db.update_ctx(ctx)

    Caveat:
    INSERTs + multiprocessing do not work well together.
    Do INSERTs in main thread.
    """

    @typechecked
    def __init__(self, ctx: Context) -> None:
        self.ctx = ctx

        # file db
        self.location_file_db = ctx.paths["database"]
        self.file_con = sqlite3.connect(self.location_file_db, uri=True)

        # memory db
        self.location_memory_db = "mem1?mode=memory&cache=shared"
        self.memory_con = sqlite3.connect(f"file:{self.location_memory_db}", uri=True)

        # copy and persist
        self.file_con.backup(self.memory_con)
        self.ctx.paths["database"] = self.location_memory_db

        self.file_con.close()  # close file db but keep memory db open

    @typechecked
    def update_ctx(self, ctx: Context) -> Context:
        """
        replace path to file-based DB with path to in-memory DB
        """
        ctx.paths["database"] = self.location_memory_db
        return ctx

    def close(self) -> None:
        self.file_con = sqlite3.connect(self.location_file_db, uri=True)
        self.memory_con.backup(self.file_con)
        self.memory_con.close()
        self.file_con.close()
        log.info("memory DB copied to file again")


@typechecked
def use_inmemory(ctx: Context) -> MemoryDB:
    """
    Creates an in-memory database and returns a MemoryDB object.

    see usage example in MemoryDB docstring

    Args:
        ctx (storagetools.Context): The context object containing the paths to the database.

    Returns:
        A MemoryDB object.
    """
    db = MemoryDB(ctx)
    log.info(f"using {db.ctx.paths['database']} as DB now")
    return db



#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Main class model of storagetools."""

import logging
import uuid
from typing import Any

from storagetools.context import Context
import storagetools.database
from typeguard import typechecked

log = logging.getLogger(__name__)


class Identifiable:
    """
    A base class that makes things identifiable with a GUID.
    """

    def __init__(self) -> None:
        self.guid = uuid.uuid4()

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        guid = f"self.guid = {self.guid.__repr__()}"
        extended_info = f"{base_repr}\n{guid}"
        return extended_info

    def supers(self) -> list[str]:
        """
        Retrieves the inheritence hierarchy of the class

        Returns:
            list of names in the method resolution order
        """
        supers = [x.__name__ for x in self.__class__.__mro__]
        return supers


class Linkable(Identifiable):
    """
    A base class that makes things linkable via attachments and associations.
    """

    def __init__(self) -> None:
        Identifiable.__init__(self)
        self.associations = []
        self.attachments = []

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        assoc_str = self.associations.__repr__()
        attac_str = self.attachments.__repr__()
        associations = f"\nself.associations = {assoc_str}"
        attachments = f"\nself.attachments = {attac_str}"
        extended_info = f"{base_repr}{associations}{attachments}"
        return extended_info

    @typechecked
    def attach_to(self, target: "Linkable") -> None:
        """
        attaches object to another object (target is root, self is leaf)
        """
        target.attachments.append(self.guid)
        self.associations.append(target.guid)

    @typechecked
    def associate(self, target: "Linkable") -> None:
        """
        associates object to with another object (target is leaf, self is root)
        """
        target.associations.append(self.guid)
        self.attachments.append(target.guid)


class HasPropertiesFromCase:
    """
    A base class that makes things have properties.
    """

    @typechecked
    def __init__(self, properties={}) -> None:
        self.properties = properties

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        extended_info = f"{base_repr}"
        for k, v in self.properties.items():
            extended_info += f"\nself.properties['{k}'] = {v}"
        return extended_info


class IsCapture:
    """
    A base class that makes things a capture of sth else.

    Intended mainly for channel and sensor captures.
    """

    @typechecked
    def __init__(self, capture_of: uuid.UUID | Identifiable) -> None:
        if isinstance(capture_of, uuid.UUID):
            self.capture_of = capture_of
        else:
            self.capture_of = capture_of.guid

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        extended_info = f"{base_repr}\nself.capture_of = {self.capture_of}"
        return extended_info


class Experiment(Linkable, HasPropertiesFromCase):
    """
    A class that represents an experiment.

    An experiment is a concept, it is atomic and defined by location, time and
    participating entities. Some practical examples: an experiment can be a
    process, a group of processes or a measurement. Basically anything that
    would get a label in a datamining context.
    """

    @typechecked
    def __init__(self, properties: dict = {}) -> None:
        Linkable.__init__(self)
        HasPropertiesFromCase.__init__(self, properties)


class Sensor(Linkable, HasPropertiesFromCase):
    """
    A class that represents a sensor.
    """

    @typechecked
    def __init__(
        self,
        properties: dict = {},
    ) -> None:
        Linkable.__init__(self)
        HasPropertiesFromCase.__init__(self, properties)


class Channel(Linkable, HasPropertiesFromCase):
    """
    A class that represents a sensor channel.
    """

    @typechecked
    def __init__(
        self,
        properties: dict = {},
    ) -> None:
        Linkable.__init__(self)
        HasPropertiesFromCase.__init__(self, properties)


class SensorCapture(Linkable, IsCapture):
    """
    A class that represents a capture of a sensor.
    """

    def __init__(self, capture_of: Identifiable) -> None:
        Linkable.__init__(self)
        IsCapture.__init__(self, capture_of)


class ChannelCapture(Linkable, IsCapture):
    """
    A class that represents a capture of a sensor channel.
    """

    def __init__(self, capture_of: Identifiable) -> None:
        Linkable.__init__(self)
        IsCapture.__init__(self, capture_of)


class TSRepresentation(Linkable):
    """
    A class that represents a time series representation.

    Time series can be representated in many ways:
    * as an array of measurement values
    * as a feature set
    * as a spectrum
    * as a series of symbols
    etc.

    In addition, time series representations can be stored in different formats:
    * numpy.Array
    * pandas.Series
    * pandas.DataFrame
    * CSV-files
    * proprietary
    etc.

    In this context, a TSRepresentation always has
    * a name, e.g., "filtered_signal"
    * a type, e.g., "pd.Series"
    * a payload that contains any object
    """

    @typechecked
    def __init__(
        self,
        payload: Any,
        name: str = "original",
        type_: str = "pd.Series",
    ) -> None:
        Linkable.__init__(self)
        self.payload = payload
        self.name = name
        self.type = type_

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        name = f"\nself.name = '{self.name}'"
        type_ = f"\nself.type= '{self.type}'"
        payload = f"\nself.payload = {self.payload.__repr__()}"
        extended_info = f"{base_repr}{name}{type_}{payload}"
        return extended_info


class Container(Linkable):
    """
    A Grouping element.

    Can hold GUIDs and optionally the deserialized objects identified by the GUIDs.

    Attributes:
        name: short_name
        info: description of what it contains or sql_statement
        lading: list that may store the deserialized attachments
    """

    @typechecked
    def __init__(self, name: str = "", info: str = "") -> None:
        super().__init__()
        if " " in name:
            # mkview will not work then...
            raise ValueError("container names may not have spaces, use _ or -")
        self.name = name
        self.info = info
        self.lading = []

    def __repr__(self) -> str:
        base_repr = super().__repr__()
        name = f"\nself.name = '{self.name}'"
        info = f"\nself.info = '{self.info}'"
        lading = f"\nself.lading = {self.lading.__repr__()}"
        extended_info = f"{base_repr}{name}{info}{lading}"
        return extended_info

    @typechecked
    def mkview(self, ctx: Context) -> None:
        """
        Creates a view in the metadata DB containing the related objects.

        Can be handy for more complex sql filterings.
        """
        sql = f"""
        CREATE VIEW IF NOT EXISTS '{self.name}' AS
            SELECT a.child AS guid
            FROM association a
            JOIN container c ON a.parent = c.guid
            WHERE c.name = '{self.name}'
        """
        with storagetools.database.Database(ctx) as db:
            db.cur.execute(sql)

    @typechecked
    def rmview(self, ctx: Context) -> None:
        """
        Removes a related view in the metadata DB if exists.
        """
        sql = f"DROP VIEW IF EXISTS {self.name}"
        with storagetools.database.Database(ctx) as db:
            db.cur.execute(sql)



#
# This file is part of storagetools.
#
# Copyright (c) 2025 Mauritz Mälzer
#
# BSD-3-Clause License
#

"""Convenience functions for working with storagetools."""

import logging
import os
import pickle
from typing import Any
from uuid import UUID, uuid4
import copy

from storagetools import database
from storagetools import db_interface as dbi
from storagetools import model
from storagetools.context import Context
from typeguard import typechecked

log = logging.getLogger(__name__)


@typechecked
def pickl(ctx: Context, target: Any, subdir: str, name: str) -> None:
    """
    Convenience function to pickle away objects in the worktree.

    Args:
        ctx: The context object containing the paths to the LMDB key-value store.
        target: thing to be pickled
        subdir: directory below "workdir" in wich to store pickle
        name: filename
    """
    os.makedirs(os.path.join(ctx.paths["work"], subdir), exist_ok=True)
    with open(os.path.join(ctx.paths["work"], subdir, name), "wb") as f:
        pickle.dump(target, f, protocol=pickle.HIGHEST_PROTOCOL)


@typechecked
def unpickl(ctx: Context, subdir: str, name: str) -> Any:
    """
    Convenience function to unpickle objects in the worktree.

    Args:
        ctx: The context object containing the paths to the LMDB key-value store.
        subdir: directory below "workdir" in wich to store pickle
        name: filename

    Returns:
        unpickled object
    """
    with open(os.path.join(ctx.paths["work"], subdir, name), "rb") as f:
        result = pickle.load(f)
    return result


@typechecked
def fetch_container_by_name(ctx: Context, name: str) -> model.Container:
    """
    Fetches container by name.

    Args:
        ctx: context
        name: container name

    Returns:
        Container
    """
    sql = f"SELECT guid FROM container WHERE name = '{name}'"
    with database.Database(ctx) as db:
        id_ = db.con.execute(sql).fetchone()[0]

    container = dbi.deserialize(ctx, id_)
    return container


@typechecked
def fetch_single_by_sql(ctx: Context, sql: str) -> Any:
    """
    Fetches first result from an SQL query.

    Note: query for GUIDs

    Args:
        ctx: context
        sql: query string for the metadata DB

    Returns:
        deserialized object that corresponds to the first sql result.
    """
    with database.Database(ctx) as db:
        id_ = db.con.execute(sql).fetchone()[0]

    x = dbi.deserialize(ctx, id_)
    return x


@typechecked
def create_container_from_guid_list(
    ctx: Context,
    guid_list: list[UUID],
    name: str = "",
    info: str = "",
) -> model.Container:
    """
    Creates a container with lading from a list of guids.

    How it works:
    in order to properly attach stuff, guids need to be deserialized first
    * deserialize all
    * attach lading to container

    Args:
        ctx: context
        guid_list: list of single-item-tuples containing guids
            (or list of guids)
        name: container name
        info: container info

    Returns:
        container instance with lading
    """
    lading = dbi.deserialize_guid_list(ctx, guid_list)

    container = model.Container(name=name, info=info)
    for thing in lading:
        thing.attach_to(container)  # type: ignore
    container.lading = lading
    return container


@typechecked
def create_container_from_sql(
    ctx: Context,
    sql: str,
    name: str = "",
    info: str = "",
) -> model.Container:
    """
    Convenience function to avoid 'with db' statement.

    Args:
        ctx: context
        sql: sql-query string that selects only guids
        name: container name
        info: container info

    Returns:
        container instance with lading
    """
    with database.Database(ctx) as db:
        guids = db.con.execute(sql).fetchall()

    guid_list = [x[0] for x in guids]
    container = create_container_from_guid_list(ctx, guid_list, name=name, info=info)
    return container


@typechecked
def deserialize_lading(ctx: Context, container: model.Container) -> None:
    """
    Deserializes attachments and stores objects into lading.
    """
    # guid_list = [(x,) for x in container.attachments]
    guid_list = container.attachments
    container.lading = dbi.deserialize_guid_list(ctx, guid_list)


@typechecked
def serialize_lading(ctx: Context, container: model.Container) -> None:
    """
    Serializes lading into DB.
    """
    dbi.serialize_list(ctx, container.lading)


@typechecked
def clone(target: model.Identifiable) -> model.Identifiable:
    """
    Creates a object with identical properties and new guid.

    Args:
        target: object to clone

    Returns:
        An Identifiable with identical properties and new guid
    """
    new = copy.deepcopy(target)
    new.guid = uuid4()
    return new
